# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Sequence, Optional, Any
from logging import Logger
import argparse
import pprint

from pyspark.sql import SparkSession

from dataproc_templates import BaseTemplate
from dataproc_templates.util.argument_parsing import add_spark_options
from dataproc_templates.util.dataframe_reader_wrappers import ingest_dataframe_from_cloud_storage
import dataproc_templates.util.template_constants as constants


__all__ = ['GCSToBigTableTemplate']


class GCSToBigTableTemplate(BaseTemplate):
    """
    Dataproc template implementing loads from GCS into BigTable
    """

    @staticmethod
    def parse_args(args: Optional[Sequence[str]] = None) -> Dict[str, Any]:
        parser: argparse.ArgumentParser = argparse.ArgumentParser()

        parser.add_argument(
            f'--{constants.GCS_BT_INPUT_LOCATION}',
            dest=constants.GCS_BT_INPUT_LOCATION,
            required=True,
            help='Cloud Storage location of the input files'
        )
        parser.add_argument(
            f'--{constants.GCS_BT_INPUT_FORMAT}',
            dest=constants.GCS_BT_INPUT_FORMAT,
            required=True,
            help='Input file format (one of: avro,parquet,csv,json,delta)',
            choices=[
                constants.FORMAT_AVRO,
                constants.FORMAT_PRQT,
                constants.FORMAT_CSV,
                constants.FORMAT_JSON,
                constants.FORMAT_DELTA
            ]
        )
        add_spark_options(parser, constants.get_csv_input_spark_options("gcs.bigtable.input."))
        parser.add_argument(
            f'--{constants.GCS_BT_HBASE_CATALOG_JSON}',
            dest=constants.GCS_BT_HBASE_CATALOG_JSON,
            required=True,
            help='HBase catalog inline json'
        )

        known_args: argparse.Namespace
        known_args, _ = parser.parse_known_args(args)

        return vars(known_args)

    def run(self, spark: SparkSession, args: Dict[str, Any]) -> None:

        logger: Logger = self.get_logger(spark=spark)

        # Arguments
        input_location: str = args[constants.GCS_BT_INPUT_LOCATION]
        input_format: str = args[constants.GCS_BT_INPUT_FORMAT]
        catalog: str = ''.join(args[constants.GCS_BT_HBASE_CATALOG_JSON].split())

        logger.info(
            "Starting Cloud Storage to BigTable Spark job with parameters:\n"
            f"{pprint.pformat(args)}"
        )

        # Read
        input_data = ingest_dataframe_from_cloud_storage(
            spark, args, input_location, input_format, "gcs.bigtable.input."
        )

        # Write
        input_data.write \
            .format(constants.FORMAT_HBASE) \
            .options(catalog=catalog) \
            .option('hbase.spark.use.hbasecontext', "false") \
            .save()



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from pyspark.sql import DataFrame, SparkSession

import dataproc_templates.util.template_constants as constants


def ingest_dataframe_from_cloud_storage(
    spark: SparkSession,
    args: dict,
    input_location: str,
    input_format: str,
    prefix: str,
    avro_format_override: Optional[str] = None,
) -> DataFrame:
    """Return a Dataframe reader object with methods and options applied for reading from Cloud Storage."""
    input_data: DataFrame

    csv_input_constant_options: dict = constants.get_csv_input_spark_options(prefix)
    spark_options = {csv_input_constant_options[k]: v
                     for k, v in args.items()
                     if k in csv_input_constant_options and v}

    if input_format == constants.FORMAT_PRQT:
        input_data = spark.read \
            .parquet(input_location)
    elif input_format == constants.FORMAT_AVRO:
        input_data = spark.read \
            .format(avro_format_override or constants.FORMAT_AVRO_EXTD) \
            .load(input_location)
    elif input_format == constants.FORMAT_CSV:
        input_data = spark.read \
            .format(constants.FORMAT_CSV) \
            .options(**spark_options) \
            .load(input_location)
    elif input_format == constants.FORMAT_JSON:
        input_data = spark.read \
            .json(input_location)
    elif input_format == constants.FORMAT_DELTA:
        input_data = spark.read \
            .format(constants.FORMAT_DELTA) \
            .load(input_location)

    return input_data



"""
 * Copyright 2022 Google LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * https://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
"""


import mock
import pyspark
from dataproc_templates.hive.util.hive_ddl_extractor import \
    HiveDDLExtractorTemplate
from datetime import datetime

class TestHiveDDLExtractorTemplate:
    """
    Test suite for HiveDDLExtractorTemplate
    """

    def test_parse_args(self):
        """Tests HiveDDLExtractorTemplate.parse_args()"""

        hive_ddl_extractor_template = HiveDDLExtractorTemplate()
        parsed_args = hive_ddl_extractor_template.parse_args(
            ["--hive.ddl.extractor.input.database=database",
             "--hive.ddl.extractor.output.path=gs://bucket/path"])

        assert parsed_args["hive.ddl.extractor.input.database"] == "database"
        assert parsed_args["hive.ddl.extractor.output.path"] == "gs://bucket/path"

    @mock.patch.object(pyspark.sql, 'SparkSession')
    def test_run(self, mock_spark_session):
        """Tests HiveDDLExtractorTemplate runs with append mode"""

        hive_ddl_extractor_template = HiveDDLExtractorTemplate()
        mock_parsed_args = hive_ddl_extractor_template.parse_args(
            ["--hive.ddl.extractor.input.database=database",
             "--hive.ddl.extractor.output.path=gs://bucket/path"])
        mock_spark_session.sql.return_value=mock_spark_session.dataframe.DataFrame
        mock_spark_session.sparkContext.parallelize.return_value=mock_spark_session.rdd.RDD
        hive_ddl_extractor_template.run(mock_spark_session, mock_parsed_args)
        mock_spark_session.sql.assert_called_once_with("SHOW TABLES IN database")
        mock_spark_session.sparkContext.parallelize.assert_called_once_with([])
        ct = datetime.now().strftime("%m-%d-%Y %H.%M.%S")
        mock_spark_session.rdd.RDD.coalesce().saveAsTextFile.assert_called_once_with("gs://bucket/path/database/{ct}".format(ct=ct))


"""
 * Copyright 2022 Google LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * https://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
"""

import mock
import pyspark
from dataproc_templates.pubsublite.pubsublite_to_gcs import PubSubLiteToGCSTemplate
import dataproc_templates.util.template_constants as constants


class TestPubSubLiteToGCSTemplate:
    """
    Test suite for PubSubLiteToGCSTemplate
    """

    def test_parse_args1(self):
        """Tests PubSubLiteToGCSTemplate.parse_args()"""

        pubsublite_to_gcs_template = PubSubLiteToGCSTemplate()
        parsed_args = pubsublite_to_gcs_template.parse_args(
            ["--pubsublite.to.gcs.input.subscription.url=test-sub",
             "--pubsublite.to.gcs.write.mode=append",
             "--pubsublite.to.gcs.output.location=gs://test",
             "--pubsublite.to.gcs.checkpoint.location=gs://test-checkpoint",
             "--pubsublite.to.gcs.output.format=csv",
             "--pubsublite.to.gcs.timeout=120",
             "--pubsublite.to.gcs.processing.time=1"
             ])

        assert parsed_args["pubsublite.to.gcs.input.subscription.url"] == "test-sub"
        assert parsed_args["pubsublite.to.gcs.write.mode"] == "append"
        assert parsed_args["pubsublite.to.gcs.output.location"] == "gs://test"
        assert parsed_args["pubsublite.to.gcs.checkpoint.location"] == "gs://test-checkpoint"
        assert parsed_args["pubsublite.to.gcs.output.format"] == "csv"
        assert parsed_args["pubsublite.to.gcs.timeout"] == 120
        assert parsed_args["pubsublite.to.gcs.processing.time"] == "1"

    def test_parse_args2(self):
        """Tests PubSubLiteToGCSTemplate.parse_args() when output format not passed"""

        pubsublite_to_gcs_template = PubSubLiteToGCSTemplate()
        parsed_args = pubsublite_to_gcs_template.parse_args(
            ["--pubsublite.to.gcs.input.subscription.url=test-sub",
             "--pubsublite.to.gcs.write.mode=append",
             "--pubsublite.to.gcs.output.location=gs://test",
             "--pubsublite.to.gcs.checkpoint.location=gs://test-checkpoint",
             "--pubsublite.to.gcs.timeout=120",
             "--pubsublite.to.gcs.processing.time=1"
             ])

        assert parsed_args["pubsublite.to.gcs.input.subscription.url"] == "test-sub"
        assert parsed_args["pubsublite.to.gcs.write.mode"] == "append"
        assert parsed_args["pubsublite.to.gcs.output.location"] == "gs://test"
        assert parsed_args["pubsublite.to.gcs.checkpoint.location"] == "gs://test-checkpoint"
        assert parsed_args["pubsublite.to.gcs.output.format"] == "csv"
        assert parsed_args["pubsublite.to.gcs.timeout"] == 120
        assert parsed_args["pubsublite.to.gcs.processing.time"] == "1"

    def test_parse_args3(self):
        """Tests PubSubLiteToGCSTemplate.parse_args() when output mode not passed"""

        pubsublite_to_gcs_template = PubSubLiteToGCSTemplate()
        parsed_args = pubsublite_to_gcs_template.parse_args(
            ["--pubsublite.to.gcs.input.subscription.url=test-sub",
             "--pubsublite.to.gcs.output.location=gs://test",
             "--pubsublite.to.gcs.checkpoint.location=gs://test-checkpoint",
             "--pubsublite.to.gcs.output.format=csv",
             "--pubsublite.to.gcs.timeout=120",
             "--pubsublite.to.gcs.processing.time=1"
             ])

        assert parsed_args["pubsublite.to.gcs.input.subscription.url"] == "test-sub"
        assert parsed_args["pubsublite.to.gcs.write.mode"] == "append"
        assert parsed_args["pubsublite.to.gcs.output.location"] == "gs://test"
        assert parsed_args["pubsublite.to.gcs.checkpoint.location"] == "gs://test-checkpoint"
        assert parsed_args["pubsublite.to.gcs.output.format"] == "csv"
        assert parsed_args["pubsublite.to.gcs.timeout"] == 120
        assert parsed_args["pubsublite.to.gcs.processing.time"] == "1"



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Sequence, Optional, Any
from logging import Logger
import argparse
import pprint

from pyspark.sql import SparkSession, DataFrame, DataFrameWriter

from dataproc_templates import BaseTemplate
import dataproc_templates.util.template_constants as constants

__all__ = ['JDBCToBigQueryTemplate']


class JDBCToBigQueryTemplate(BaseTemplate):
    """
    Dataproc template implementing loads from JDBC into BigQuery
    """

    @staticmethod
    def parse_args(args: Optional[Sequence[str]] = None) -> Dict[str, Any]:
        parser: argparse.ArgumentParser = argparse.ArgumentParser()

        parser.add_argument(
            f'--{constants.JDBC_BQ_OUTPUT_DATASET}',
            dest=constants.JDBC_BQ_OUTPUT_DATASET,
            required=True,
            help='BigQuery dataset for the output table'
        )
        parser.add_argument(
            f'--{constants.JDBC_BQ_OUTPUT_TABLE}',
            dest=constants.JDBC_BQ_OUTPUT_TABLE,
            required=True,
            help='BigQuery output table name'
        )
        parser.add_argument(
            f'--{constants.JDBC_BQ_LD_TEMP_BUCKET_NAME}',
            dest=constants.JDBC_BQ_LD_TEMP_BUCKET_NAME,
            required=True,
            help='Spark BigQuery connector temporary bucket'
        )
        parser.add_argument(
            f'--{constants.JDBC_BQ_INPUT_URL}',
            dest=constants.JDBC_BQ_INPUT_URL,
            required=True,
            help='JDBC input URL'
        )
        parser.add_argument(
            f'--{constants.JDBC_BQ_INPUT_DRIVER}',
            dest=constants.JDBC_BQ_INPUT_DRIVER,
            required=True,
            help='JDBC input driver name'
        )
        parser.add_argument(
            f'--{constants.JDBC_BQ_INPUT_TABLE}',
            dest=constants.JDBC_BQ_INPUT_TABLE,
            required=True,
            help='JDBC input table name'
        )
        parser.add_argument(
            f'--{constants.JDBC_BQ_INPUT_PARTITIONCOLUMN}',
            dest=constants.JDBC_BQ_INPUT_PARTITIONCOLUMN,
            required=False,
            default="",
            help='JDBC input table partition column name'
        )
        parser.add_argument(
            f'--{constants.JDBC_BQ_INPUT_LOWERBOUND}',
            dest=constants.JDBC_BQ_INPUT_LOWERBOUND,
            required=False,
            default="",
            help='JDBC input table partition column lower bound which is used to decide the partition stride'
        )
        parser.add_argument(
            f'--{constants.JDBC_BQ_INPUT_UPPERBOUND}',
            dest=constants.JDBC_BQ_INPUT_UPPERBOUND,
            required=False,
            default="",
            help='JDBC input table partition column upper bound which is used to decide the partition stride'
        )
        parser.add_argument(
            f'--{constants.JDBC_BQ_NUMPARTITIONS}',
            dest=constants.JDBC_BQ_NUMPARTITIONS,
            required=False,
            default="10",
            help='The maximum number of partitions that can be used for parallelism in table reading and writing. Default set to 10'
        )
        parser.add_argument(
            f'--{constants.JDBC_BQ_INPUT_FETCHSIZE}',
            dest=constants.JDBC_BQ_INPUT_FETCHSIZE,
            required=False,
            default=0,
            type=int,
            help='Determines how many rows to fetch per round trip'
        )
        parser.add_argument(
            f'--{constants.JDBC_BQ_SESSIONINITSTATEMENT}',
            dest=constants.JDBC_BQ_SESSIONINITSTATEMENT,
            required=False,
            default="",
            help='Custom SQL statement to execute in each reader database session'
        )
        parser.add_argument(
            f'--{constants.JDBC_BQ_OUTPUT_MODE}',
            dest=constants.JDBC_BQ_OUTPUT_MODE,
            required=False,
            default=constants.OUTPUT_MODE_APPEND,
            help=(
                'Output write mode '
                '(one of: append,overwrite,ignore,errorifexists) '
                '(Defaults to append)'
            ),
            choices=[
                constants.OUTPUT_MODE_OVERWRITE,
                constants.OUTPUT_MODE_APPEND,
                constants.OUTPUT_MODE_IGNORE,
                constants.OUTPUT_MODE_ERRORIFEXISTS
            ]
        )

        known_args: argparse.Namespace
        known_args, _ = parser.parse_known_args(args)

        return vars(known_args)

    def run(self, spark: SparkSession, args: Dict[str, Any]) -> None:

        logger: Logger = self.get_logger(spark=spark)

        # Arguments
        big_query_dataset: str = args[constants.JDBC_BQ_OUTPUT_DATASET]
        big_query_table: str = args[constants.JDBC_BQ_OUTPUT_TABLE]
        bq_temp_bucket: str = args[constants.JDBC_BQ_LD_TEMP_BUCKET_NAME]
        input_jdbc_url: str = args[constants.JDBC_BQ_INPUT_URL]
        input_jdbc_driver: str = args[constants.JDBC_BQ_INPUT_DRIVER]
        input_jdbc_table: str = args[constants.JDBC_BQ_INPUT_TABLE]
        input_jdbc_partitioncolumn: str = args[constants.JDBC_BQ_INPUT_PARTITIONCOLUMN]
        input_jdbc_lowerbound: str = args[constants.JDBC_BQ_INPUT_LOWERBOUND]
        input_jdbc_upperbound: str = args[constants.JDBC_BQ_INPUT_UPPERBOUND]
        jdbc_numpartitions: str = args[constants.JDBC_BQ_NUMPARTITIONS]
        input_jdbc_fetchsize: int = args[constants.JDBC_BQ_INPUT_FETCHSIZE]
        input_jdbc_sessioninitstatement: str = args[constants.JDBC_BQ_SESSIONINITSTATEMENT]
        output_mode: str = args[constants.JDBC_BQ_OUTPUT_MODE]

        logger.info(
            "Starting JDBC to BigQuery Spark job with parameters:\n"
            f"{pprint.pformat(args)}"
        )

        # Read
        input_data: DataFrame

        partition_parameters = str(input_jdbc_partitioncolumn) + str(input_jdbc_lowerbound) + str(input_jdbc_upperbound)
        if ((partition_parameters != "") & ((input_jdbc_partitioncolumn == "") | (input_jdbc_lowerbound == "") | (input_jdbc_upperbound == ""))):
            logger.error("Set all the sql partitioning parameters together-jdbctogcs.input.partitioncolumn,jdbctogcs.input.lowerbound,jdbctogcs.input.upperbound. Refer to README.md for more instructions.")
            exit (1)

        properties = {constants.JDBC_URL: input_jdbc_url,
                      constants.JDBC_DRIVER: input_jdbc_driver,
                      constants.JDBC_TABLE: input_jdbc_table,
                      constants.JDBC_NUMPARTITIONS: jdbc_numpartitions,
                      constants.JDBC_FETCHSIZE: input_jdbc_fetchsize}
        if input_jdbc_sessioninitstatement:
            properties[constants.JDBC_SESSIONINITSTATEMENT] = input_jdbc_sessioninitstatement
        if partition_parameters:
            properties.update({constants.JDBC_PARTITIONCOLUMN: input_jdbc_partitioncolumn,
                               constants.JDBC_LOWERBOUND: input_jdbc_lowerbound,
                               constants.JDBC_UPPERBOUND: input_jdbc_upperbound})

        input_data = spark.read \
            .format(constants.FORMAT_JDBC) \
            .options(**properties) \
            .load()

        # Write
        input_data.write \
                .format(constants.FORMAT_BIGQUERY) \
                .option(constants.TABLE, big_query_dataset + "." + big_query_table) \
                .option(constants.GCS_BQ_TEMP_BUCKET, bq_temp_bucket) \
                .mode(output_mode) \
                .save()




# -*- coding: utf-8 -*-
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Setup script for the dataproc_templates package

It extends the bdist_egg command to allow renaming of the output .egg artifact.
Usage: `python setup.py bdist_egg --output=my-custom-artifact-name.egg`
"""

from typing import Optional
import shutil
import io
import os
from distutils import log
from setuptools import setup, find_packages
from setuptools.command.bdist_egg import bdist_egg


class BdistEggCustomEggName(bdist_egg):
    """
    Create an "egg" distrubution, optionally renaming the output .egg file

    This custom command extends the bdist_egg command. It adds the
    `--output` argument for specifying the name of the output .egg file.
    If it's provided, the egg artifact is renamed after it's built.
    """

    description: str = 'create an "egg" distrubution, optionally renaming the output .egg file'

    user_options = [('output=', 'o', 'output file name')] + \
        bdist_egg.user_options

    def __init__(self, *args, **kwargs):
        self.output: Optional[str] = None
        super().__init__(*args, **kwargs)

    def initialize_options(self):
        """Initialize command options"""
        super().initialize_options()
        self.output = None

    def run(self):
        """
        Run the command

        First execute `bdist_egg` as usual. Then, rename the .egg file
        if the `--output` argument was specified.
        """

        super().run()

        built_egg_path: str = self.get_outputs()[0]

        if self.output is not None:
            log.info(
                'Will rename output .egg file from %s to %s',
                built_egg_path,
                self.output
            )
            shutil.move(src=built_egg_path, dst=self.output)


dependencies = [
    "pyspark>=3.2.0",
    "google-cloud-bigquery>=3.4.0"
]


package_root = os.path.abspath(os.path.dirname(__file__))

version = {}
with open(
    os.path.join(package_root, "version.py")
) as fp:
    exec(fp.read(), version)
version = version["__version__"]

readme_filename = os.path.join(package_root, "README.md")
with io.open(readme_filename, encoding="utf-8") as readme_file:
    readme = readme_file.read()

setup(
    name="google-dataproc-templates",
    version=version,
    description="Google Dataproc templates written in Python",
    long_description_content_type="text/markdown",
    long_description=readme,
    license="Apache 2.0",
    url="https://github.com/GoogleCloudPlatform/dataproc-templates",
    packages=find_packages(exclude=['test']),
    install_requires=dependencies,
    # Override the bdist_egg command with our extension
    cmdclass={
        'bdist_egg': BdistEggCustomEggName
    },
    platforms="Posix; MacOS X; Windows",
    python_requires=">=3.7"
)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import List
from enum import Enum

__all__ = ['TemplateName']


class TemplateName(Enum):
    """Enumeration of all template types"""

    GCSTOBIGQUERY = "GCSTOBIGQUERY"
    GCSTOBIGTABLE = "GCSTOBIGTABLE"
    GCSTOGCS = "GCSTOGCS"
    BIGQUERYTOGCS = "BIGQUERYTOGCS"
    HIVETOBIGQUERY = "HIVETOBIGQUERY"
    HIVETOGCS = "HIVETOGCS"
    TEXTTOBIGQUERY = "TEXTTOBIGQUERY"
    GCSTOJDBC = "GCSTOJDBC"
    GCSTOMONGO = "GCSTOMONGO"
    HBASETOGCS = "HBASETOGCS"
    JDBCTOJDBC = "JDBCTOJDBC"
    JDBCTOGCS = "JDBCTOGCS"
    JDBCTOBIGQUERY = "JDBCTOBIGQUERY"
    MONGOTOGCS = "MONGOTOGCS"
    SNOWFLAKETOGCS = "SNOWFLAKETOGCS"
    REDSHIFTTOGCS = "REDSHIFTTOGCS"
    CASSANDRATOBQ= "CASSANDRATOBQ"
    AZUREBLOBSTORAGETOBQ = "AZUREBLOBSTORAGETOBQ"
    CASSANDRATOGCS = "CASSANDRATOGCS"
    HIVEDDLEXTRACTOR = "HIVEDDLEXTRACTOR"
    KAFKATOGCS="KAFKATOGCS"
    KAFKATOBQ="KAFKATOBQ"
    S3TOBIGQUERY = "S3TOBIGQUERY"
    PUBSUBLITETOGCS = "PUBSUBLITETOGCS"



    @classmethod
    def from_string(cls, template_name: str) -> TemplateName:
        """
        Get the TemplateName value from a string.

        This is not case-sensitive.

        Args:
            template_name (str): The string representation of
                the template name.

        Returns:
            TemplateName: the TemplateName value corresponding
                to the provided string

        Raises:
            ValueError: if the given template name is invalid
        """

        try:
            return cls[template_name.upper()]
        except KeyError as err:
            raise ValueError(f'Invalid template name {template_name}') from err

    @classmethod
    def choices(cls) -> List[str]:
        """
        Returns all the available TemplateName options as strings

        Returns:
            List[str]: All available TemplateName options
        """

        return [
            template_name.value for template_name in cls
        ]



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from pyspark.sql import DataFrame, SparkSession

import dataproc_templates.util.template_constants as constants


def persist_dataframe_to_cloud_storage(
    input_dataframe: DataFrame,
    args: dict,
    output_location: str,
    output_format: str,
    prefix: str
) -> DataFrame:
    """Persist input_dataframe object with methods and options applied for writing to Cloud Storage."""

    csv_output_constant_options: dict = constants.get_csv_output_spark_options(prefix)
    spark_options = {csv_output_constant_options[k]: v for k, v in args.items() if k in csv_output_constant_options and v}

    if output_format == constants.FORMAT_PRQT:
        input_dataframe \
            .parquet(output_location)
    elif output_format == constants.FORMAT_AVRO:
        input_dataframe \
            .format(constants.FORMAT_AVRO) \
            .save(output_location)
    elif output_format == constants.FORMAT_CSV:
        input_dataframe \
            .options(**spark_options) \
            .csv(output_location)
    elif output_format == constants.FORMAT_JSON:
        input_dataframe \
            .json(output_location)



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Sequence, Optional, Any
from logging import Logger
import argparse
import pprint
import sys

from pyspark.sql import SparkSession, DataFrame, DataFrameWriter

from dataproc_templates import BaseTemplate
import dataproc_templates.util.template_constants as constants
from dataproc_templates.util.argument_parsing import add_spark_options
from dataproc_templates.util.dataframe_reader_wrappers import ingest_dataframe_from_cloud_storage
from dataproc_templates.util.dataframe_writer_wrappers import persist_dataframe_to_cloud_storage


__all__ = ['GCSToGCSTemplate']


class GCSToGCSTemplate(BaseTemplate):
    """
    Dataproc template implementing loads from Cloud Storage into Cloud Storage post SQL transformation
    """

    @staticmethod
    def parse_args(args: Optional[Sequence[str]] = None) -> Dict[str, Any]:

        parser: argparse.ArgumentParser = argparse.ArgumentParser()

        parser.add_argument(
            f'--{constants.GCS_TO_GCS_INPUT_LOCATION}',
            dest=constants.GCS_TO_GCS_INPUT_LOCATION,
            required=True,
            help='Cloud Storage location of the input files'
        )
        parser.add_argument(
            f'--{constants.GCS_TO_GCS_INPUT_FORMAT}',
            dest=constants.GCS_TO_GCS_INPUT_FORMAT,
            required=True,
            help='Cloud Storage input file format (one of: avro,parquet,csv,json,delta)',
            choices=[
                constants.FORMAT_AVRO,
                constants.FORMAT_PRQT,
                constants.FORMAT_CSV,
                constants.FORMAT_JSON,
                constants.FORMAT_DELTA
            ]
        )
        add_spark_options(parser, constants.get_csv_input_spark_options("gcs.gcs.input."))
        add_spark_options(parser, constants.get_csv_output_spark_options("gcs.gcs.output."))
        parser.add_argument(
            f'--{constants.GCS_TO_GCS_TEMP_VIEW_NAME}',
            dest=constants.GCS_TO_GCS_TEMP_VIEW_NAME,
            required=False,
            default="",
            help='Temp view name for creating a spark sql view on source data. This name has to match with the table name that will be used in the SQL query'
        )
        parser.add_argument(
            f'--{constants.GCS_TO_GCS_SQL_QUERY}',
            dest=constants.GCS_TO_GCS_SQL_QUERY,
            required=False,
            default="",
            help='SQL query for data transformation. This must use the temp view name as the table to query from.'
        )

        parser.add_argument(
            f'--{constants.GCS_TO_GCS_OUTPUT_PARTITION_COLUMN}',
            dest=constants.GCS_TO_GCS_OUTPUT_PARTITION_COLUMN,
            required=False,
            default="",
            help='Partition column name to partition the final output in destination bucket'
        )
        parser.add_argument(
            f'--{constants.GCS_TO_GCS_OUTPUT_FORMAT}',
            dest=constants.GCS_TO_GCS_OUTPUT_FORMAT,
            required=False,
            default=constants.FORMAT_PRQT,
            help=(
                'Output write format '
                '(one of: avro,parquet,csv,json)'
                '(Defaults to parquet)'
            ),
            choices=[
                constants.FORMAT_AVRO,
                constants.FORMAT_PRQT,
                constants.FORMAT_CSV,
                constants.FORMAT_JSON
            ]
        )

        parser.add_argument(
            f'--{constants.GCS_TO_GCS_OUTPUT_MODE}',
            dest=constants.GCS_TO_GCS_OUTPUT_MODE,
            required=False,
            default=constants.OUTPUT_MODE_APPEND,
            help=(
                'Output write mode '
                '(one of: append,overwrite,ignore,errorifexists) '
                '(Defaults to append)'
            ),
            choices=[
                constants.OUTPUT_MODE_OVERWRITE,
                constants.OUTPUT_MODE_APPEND,
                constants.OUTPUT_MODE_IGNORE,
                constants.OUTPUT_MODE_ERRORIFEXISTS
            ]
        )

        parser.add_argument(
            f'--{constants.GCS_TO_GCS_OUTPUT_LOCATION}',
            dest=constants.GCS_TO_GCS_OUTPUT_LOCATION,
            required=True,
            help=(
                'Destination Cloud Storage location'
            )
        )

        known_args: argparse.Namespace
        known_args, _ = parser.parse_known_args(args)

        if getattr(known_args, constants.GCS_TO_GCS_SQL_QUERY) and not getattr(known_args, constants.GCS_TO_GCS_TEMP_VIEW_NAME):
            sys.exit('ArgumentParser Error: Temp view name cannot be null if you want to do data transformations with query')

        return vars(known_args)

    def run(self, spark: SparkSession, args: Dict[str, Any]) -> None:

        logger: Logger = self.get_logger(spark=spark)
        # Arguments
        input_location: str = args[constants.GCS_TO_GCS_INPUT_LOCATION]
        input_format: str = args[constants.GCS_TO_GCS_INPUT_FORMAT]
        gcs_temp_view: str = args[constants.GCS_TO_GCS_TEMP_VIEW_NAME]
        sql_query: str = args[constants.GCS_TO_GCS_SQL_QUERY]
        output_partition_column: str = args[constants.GCS_TO_GCS_OUTPUT_PARTITION_COLUMN]
        output_mode: str = args[constants.GCS_TO_GCS_OUTPUT_MODE]
        output_format: str = args[constants.GCS_TO_GCS_OUTPUT_FORMAT]
        output_location: str = args[constants.GCS_TO_GCS_OUTPUT_LOCATION]

        logger.info(
            "Starting Cloud Storage to Cloud Storage with tranformations Spark job with parameters:\n"
            f"{pprint.pformat(args)}"
        )

        # Read
        input_data = ingest_dataframe_from_cloud_storage(
            spark,
            args,
            input_location,
            input_format,
            "gcs.gcs.input.",
            avro_format_override=constants.FORMAT_AVRO
        )

        if sql_query:
            # Create temp view on source data
            input_data.createOrReplaceTempView(gcs_temp_view)
            # Execute SQL
            output_data = spark.sql(sql_query)
        else:
            output_data = input_data

        # Write
        if output_partition_column:
            writer: DataFrameWriter = output_data.write.mode(output_mode).partitionBy(output_partition_column)
        else:
            writer: DataFrameWriter = output_data.write.mode(output_mode)

        persist_dataframe_to_cloud_storage(writer, args, output_location, output_format, "gcs.gcs.output.")


"""
 * Copyright 2022 Google LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * https://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
"""

import mock
import pyspark

from dataproc_templates.gcs.gcs_to_bigtable import GCSToBigTableTemplate
import dataproc_templates.util.template_constants as constants


class TestGCSToBigTableTemplate:
    """
    Test suite for GCSToBigTableTemplate
    """

    def test_parse_args(self):
        """Tests GCSToBigTableTemplate.parse_args()"""

        gcs_to_bigtable_template = GCSToBigTableTemplate()
        parsed_args = gcs_to_bigtable_template.parse_args(
            ["--gcs.bigtable.input.format=parquet",
             "--gcs.bigtable.input.location=gs://test",
             "--gcs.bigtable.hbase.catalog.json={key:value}"])

        assert parsed_args["gcs.bigtable.input.format"] == "parquet"
        assert parsed_args["gcs.bigtable.input.location"] == "gs://test"
        assert parsed_args["gcs.bigtable.hbase.catalog.json"] == '{key:value}'

    @mock.patch.object(pyspark.sql, 'SparkSession')
    def test_run(self, mock_spark_session):
        """Tests GCSToBigTableTemplate runs"""

        gcs_to_bigtable_template = GCSToBigTableTemplate()
        mock_parsed_args = gcs_to_bigtable_template.parse_args(
            ["--gcs.bigtable.input.format=parquet",
             "--gcs.bigtable.input.location=gs://test",
             "--gcs.bigtable.hbase.catalog.json={key:value}"])
        mock_spark_session.read.parquet.return_value = mock_spark_session.dataframe.DataFrame
        gcs_to_bigtable_template.run(mock_spark_session, mock_parsed_args)

        mock_spark_session.read.parquet.assert_called_once_with("gs://test")
        mock_spark_session.dataframe.DataFrame.write.format. \
            assert_called_once_with(constants.FORMAT_HBASE)
        mock_spark_session.dataframe.DataFrame.write.format().options. \
            assert_called_with(catalog='{key:value}')
        mock_spark_session.dataframe.DataFrame.write.format().options().option. \
            assert_called_once_with('hbase.spark.use.hbasecontext', "false")

    @mock.patch.object(pyspark.sql, 'SparkSession')
    def test_run_csv1(self, mock_spark_session):
        """Tests GCSToBigTableTemplate runs with csv format"""

        gcs_to_bigtable_template = GCSToBigTableTemplate()
        mock_parsed_args = gcs_to_bigtable_template.parse_args(
            ["--gcs.bigtable.input.format=csv",
             "--gcs.bigtable.input.location=gs://test",
             "--gcs.bigtable.input.header=false",
             "--gcs.bigtable.hbase.catalog.json={key:value}"])
        mock_spark_session.read.format().options().load.return_value = mock_spark_session.dataframe.DataFrame
        gcs_to_bigtable_template.run(mock_spark_session, mock_parsed_args)

        mock_spark_session.read.format.assert_called_with(
            constants.FORMAT_CSV)
        mock_spark_session.read.format().options.assert_called_with(**{
            constants.CSV_HEADER: 'false',
            constants.CSV_INFER_SCHEMA: 'true',
        })
        mock_spark_session.read.format().options().load.assert_called_once_with("gs://test")
        mock_spark_session.dataframe.DataFrame.write.format. \
            assert_called_once_with(constants.FORMAT_HBASE)
        mock_spark_session.dataframe.DataFrame.write.format().options. \
            assert_called_with(catalog='{key:value}')
        mock_spark_session.dataframe.DataFrame.write.format().options().option. \
            assert_called_once_with('hbase.spark.use.hbasecontext', "false")

    @mock.patch.object(pyspark.sql, 'SparkSession')
    def test_run_csv2(self, mock_spark_session):
        """Tests GCSToBigTableTemplate runs with csv format and some optional csv options"""

        gcs_to_bigtable_template = GCSToBigTableTemplate()
        mock_parsed_args = gcs_to_bigtable_template.parse_args(
            ["--gcs.bigtable.input.format=csv",
             "--gcs.bigtable.input.location=gs://test",
             "--gcs.bigtable.input.inferschema=false",
             "--gcs.bigtable.input.sep=|",
             "--gcs.bigtable.input.comment=#",
             "--gcs.bigtable.input.timestampntzformat=yyyy-MM-dd'T'HH:mm:ss",
             "--gcs.bigtable.hbase.catalog.json={key:value}"])
        mock_spark_session.read.format().options().load.return_value = mock_spark_session.dataframe.DataFrame
        gcs_to_bigtable_template.run(mock_spark_session, mock_parsed_args)

        mock_spark_session.read.format.assert_called_with(
            constants.FORMAT_CSV)
        mock_spark_session.read.format().options.assert_called_with(**{
            constants.CSV_HEADER: 'true',
            constants.CSV_INFER_SCHEMA: 'false',
            constants.CSV_SEP: "|",
            constants.CSV_COMMENT: "#",
            constants.CSV_TIMESTAMPNTZFORMAT: "yyyy-MM-dd'T'HH:mm:ss",
        })
        mock_spark_session.read.format().options().load.assert_called_once_with("gs://test")
        mock_spark_session.dataframe.DataFrame.write.format. \
            assert_called_once_with(constants.FORMAT_HBASE)
        mock_spark_session.dataframe.DataFrame.write.format().options. \
            assert_called_with(catalog='{key:value}')
        mock_spark_session.dataframe.DataFrame.write.format().options().option. \
            assert_called_once_with('hbase.spark.use.hbasecontext', "false")



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Sequence, Optional, Any
from logging import Logger
import argparse
import pprint

from pyspark.sql import SparkSession

from dataproc_templates import BaseTemplate
from dataproc_templates.util.argument_parsing import add_spark_options
from dataproc_templates.util.dataframe_reader_wrappers import ingest_dataframe_from_cloud_storage
import dataproc_templates.util.template_constants as constants


__all__ = ['TextToBigQueryTemplate']


class TextToBigQueryTemplate(BaseTemplate):
    """
    Dataproc template implementing Text loads from GCS into BigQuery
    """

    @staticmethod
    def parse_args(args: Optional[Sequence[str]] = None) -> Dict[str, Any]:
        parser: argparse.ArgumentParser = argparse.ArgumentParser()

        parser.add_argument(
            f'--{constants.TEXT_BQ_INPUT_LOCATION}',
            dest=constants.TEXT_BQ_INPUT_LOCATION,
            required=True,
            help='Cloud Storage location of the input text files'
        )
        add_spark_options(parser, constants.get_csv_input_spark_options("text.bigquery.input."))
        parser.add_argument(
            f'--{constants.TEXT_BQ_OUTPUT_DATASET}',
            dest=constants.TEXT_BQ_OUTPUT_DATASET,
            required=True,
            help='BigQuery dataset for the output table'
        )
        parser.add_argument(
            f'--{constants.TEXT_BQ_OUTPUT_TABLE}',
            dest=constants.TEXT_BQ_OUTPUT_TABLE,
            required=True,
            help='BigQuery output table name'
        )
        parser.add_argument(
            f'--{constants.TEXT_BQ_LD_TEMP_BUCKET_NAME}',
            dest=constants.TEXT_BQ_LD_TEMP_BUCKET_NAME,
            required=True,
            help='Spark BigQuery connector temporary bucket'
        )
        parser.add_argument(
            f'--{constants.TEXT_BQ_OUTPUT_MODE}',
            dest=constants.TEXT_BQ_OUTPUT_MODE,
            required=False,
            default=constants.OUTPUT_MODE_APPEND,
            help=(
                'Output write mode '
                '(one of: append,overwrite,ignore,errorifexists) '
                '(Defaults to append)'
            ),
            choices=[
                constants.OUTPUT_MODE_OVERWRITE,
                constants.OUTPUT_MODE_APPEND,
                constants.OUTPUT_MODE_IGNORE,
                constants.OUTPUT_MODE_ERRORIFEXISTS
            ]
        )
        parser.add_argument(
            f'--{constants.TEXT_INPUT_COMPRESSION}',
            dest=constants.TEXT_INPUT_COMPRESSION,
            required=True,
            help='Input file compression format (one of: bzip2,deflate,lz4,gzip,None)',
            default=None,
            choices=[
                constants.COMPRESSION_BZIP2,
                constants.COMPRESSION_GZIP,
                constants.COMPRESSION_DEFLATE,
                constants.COMPRESSION_LZ4,
                constants.COMPRESSION_NONE
            ]
        )
        parser.add_argument(
                    f'--{constants.TEXT_INPUT_DELIMITER}',
                    dest=constants.TEXT_INPUT_DELIMITER,
                    required=False,
                    help=(
                        'Input column delimiter '
                        '(example: ",", ";", "|", "/","\" ) '
                    )
                )


        known_args: argparse.Namespace
        known_args, _ = parser.parse_known_args(args)

        return vars(known_args)

    def run(self, spark: SparkSession, args: Dict[str, Any]) -> None:
        logger: Logger = self.get_logger(spark=spark)

        # Arguments
        input_location: str = args[constants.TEXT_BQ_INPUT_LOCATION]
        big_query_dataset: str = args[constants.TEXT_BQ_OUTPUT_DATASET]
        big_query_table: str = args[constants.TEXT_BQ_OUTPUT_TABLE]
        bq_temp_bucket: str = args[constants.TEXT_BQ_LD_TEMP_BUCKET_NAME]
        output_mode: str = args[constants.TEXT_BQ_OUTPUT_MODE]
        # These options are redundant but left in place until this template is removed
        # via issue https://github.com/GoogleCloudPlatform/dataproc-templates/issues/721
        # input_delimiter: str = args[constants.TEXT_INPUT_DELIMITER]
        # input_file_codec_format: str = args[constants.TEXT_INPUT_COMPRESSION]

        logger.info(
            "Starting GCS to Bigquery Spark job with parameters:\n"
            f"{pprint.pformat(args)}"
        )

        # Read
        input_data = ingest_dataframe_from_cloud_storage(
            spark,
            args,
            input_location,
            constants.FORMAT_CSV,
            "text.bigquery.input.",
        )

        # Write
        input_data.write \
            .format(constants.FORMAT_BIGQUERY) \
            .option(constants.TABLE, big_query_dataset + "." + big_query_table) \
            .option(constants.TEXT_BQ_TEMP_BUCKET, bq_temp_bucket) \
            .mode(output_mode) \
            .save()
